##############    File name   CommonOptionsFunctions.R    ##############
######################################################################
####    Various functions used for options calculations  
####    Includes
####    1>  Getting online quotes from Yahoo
####    2>  Reading in a pre-downloaded from CBOE in their format
####    3>  Profit/Loss cals
####    4>  Combinatorial analysis
####    5>  Plotting setup
####    5>  Other miscellaneous
#####################################################################
#####################################################################
initOptions <- function (name = "optTrade", parmList = parmList ) 
{
    if (exists(paste("Options", name, sep = "."), 
        envir = .options, inherits = TRUE)) 
        stop("options ", name, " already exists, use updateOptions() or create a new options.")
    options = list()
    options$id = list(name = parmList$stratName, size = parmList$lotSize)
    options$segments = parmList$optStrat
    
    assign(paste("Options", name, sep = "."), options, envir = .options)
    return(name)
}


###############
getOptions <- function (aname, envir = .options) 
{
   
    if (!grepl("Options\\.", aname)) 
        Options <- try(get(paste("Options", aname, sep = "."), 
            envir = envir), silent = TRUE)
    else Options <- try(get(aname, envir = envir), silent = TRUE)
    if (inherits(Options, "try-error")) 
        stop(paste("Options ", aname, " not found, use initOptions() to create a new Options"))
    ##if (!inherits(Options, "Options")) 
      ##  stop("Options ", aname, " passed is not the name of an Options object.")
    return(Options)
}


#################

###########################################################
###   Options in R
###   Analyze profit profile of a particular option strategy 
###   Strategy must be defined in the list data object
###
###   derived from the OptionsStrategyPL  Single case
###   Analyze multiple strike price points
###   Include also the multi-legs
###   Need to define some constraints
###   May be difficult unless its setup as external database
###
############################################################

###   Inputs needed - will use SPY as example - 
###   Strategy must be setup in the optStrat list below
###   >  legs of option with type, strike price, premium etc.,
###   > name of strategy for reference
###    will analyze 25% movements below & above current price
###########################################################


###################################################
##    !!!!    This function assumes premium is absolute / always positive
###           Formulas have to adjust -  lesser confusion in setting up strategy !!!
#########################################################
calcOptData <- function(optStrat = optStrat, stratName = stratName, atExpiry = atExpiry) {
  lstProfits = lapply(optStrat, function(x){
    
    strkPrice = x$strike
    premium = x$premium
    profit = switch( x$name,
                     "Long"      =  atExpiry - strkPrice - premium, 
                     "Short"     =  strkPrice - atExpiry - premium,
                     ###  
                     "Buy Put"   =  ifelse(atExpiry < strkPrice, strkPrice - atExpiry - premium, -premium),
                     # from (b)
                     "Buy Call"  =  ifelse(atExpiry > strkPrice, atExpiry - strkPrice - premium, -premium),
                     # from (c)
                     "Sell Put" =  ifelse(atExpiry <= strkPrice, atExpiry - strkPrice + premium, premium),
                     # from (d)
                     "Sell Call" =  ifelse(atExpiry >= strkPrice, strkPrice - atExpiry + premium,  premium)
    )
    profit * x$lotsize
    
  }  )
  ###   Now lets add up all individual option profit/loss
  dat1 = as.data.frame(lstProfits)
  colnames(dat1)  =  seq(1,ncol(dat1))
  dat1$sumProfit = (apply(dat1,1,sum))
  dat1$atExpiry = atExpiry
  return(dat1)
}
###
###     plot the payoff chart for the entire strategy
###    if we want to show strike price lines - need to pass in strike price list
ShowOptionPL <- function(dat1, stratName = stratName, mktPrice = mktPrice , specs ) {
  ####  plot total
  strikes = unlist(lapply(specs, function(x)   x$strike))
  
  netAmt = unlist( sapply(specs,  function(x)  {
                      amt = round(ifelse (str_detect  (x$name, "Buy"), -x$premium*x$lotsize, x$premium*x$lotsize))
                            }) )
  
  subt = paste0(" Strikes @ ", paste(strikes, collapse= " :: "))
  ttl = paste0(stratName,  subt, "\n",  "Premium: ", paste0( netAmt, collapse = ";"), " Net Cr/Dr ", sum(netAmt) )
  gp = ggplot(dat1, aes(x=atExpiry, y = sumProfit)) + geom_line(color="blue") 
  gp = gp + labs(x="Exercise Price", y="Aggregate Profit")    ###,   title= ttl) 
  gp = gp + theme(axis.title.x = element_text(size = rel(0.6)) )
  gp = gp + theme(axis.title.y = element_text(size = rel(0.6)) )
  gp = gp + geom_ribbon(aes(ymin=pmin(sumProfit,0), ymax=0), fill="red", col="salmon4", alpha=0.5) +
            geom_ribbon(aes(ymin=0, ymax=pmax(sumProfit,0)), fill="darkgreen", col="purple3", alpha=0.5) +
            geom_line(aes(y=0))
  gp = gp + annotate("text", x = mean(range(dat1$atExpiry)), y = Inf, label = ttl, vjust=1.0, size=3)
  gp = gp + geom_vline(xintercept = strikes, colour="red", show.legend=TRUE)
  gp = gp + geom_vline(xintercept = mktPrice, colour="darkgreen")
  gp = gp + geom_hline(yintercept = 0, colour="black", size=1.2)
  return(gp)
}

#################
###   First define the strategy  and setup the parameters
###   Then determine which legs/prices will be analyzed
###   We can probably read this from the downloaded Yahoo quotes
###   So must know expiration  -  3rd Friday routines
################

###    legs of strategy must have some additional iput
###    range of strike price range - can specify as itm, atm or otm
###    default will be to use +/- 20% for itm / otm; +/- 1% atm
###    current price is critical  -  can be outside every leg
###    premiums will have to obtained from downloaded quote
###    lotsize
###############################################################

###   Legs will be specified so; then setup vector of prices to evaluate - shouldn't be too many
###   Else payoff chart will just get too busy ( use nbrStrikes for number )
####
###################
setRTStrategy  <-  function( specStrategy ) {
  ## if item null, use default; test with is.null(obj)
  ## at strategy level; then at each individal leg level
  ##  expand grid to enumerate all combinations
  ###  dont have range for ATM thats just one; only have 2 for itm or otm
  ## nbrStrikes = ifelse( is.null(specStrategy$nbrStrikes) ,tradeDefaults$nbrStrikes, specStrategy$nbrStrikes)
  ########
  ####     if instead just fixed prices to test; will be given:
  ###      Number of possibilities; one row /possibility of prices mapping to leg
  ###      will be in the FixedStrikes matrix
  ###      setup as many optGrid rows as fixedStrike rows
  
  if (tradeDefaults$fixedVals) {    ##  used fixed strikes if TRUE - supplied in fixedStrikes
    ls = list(); for ( i in 1:length(specStrategy$legs$trades)) ls[[i]]=1
    optGrid = expand.grid ( ls )
    optGrid [ 1: length(fixedStrikes), ]  =  optGrid[1,]
    
    }  else   
    optGrid = expand.grid (lapply(specStrategy$legs$trades, function(x) { 
                if (str_detect  (x$name, "(Long)|(Short)")) q = 1 else
                  if(x$moneyX == "ATM") q =  0 else
                    if (x$moneyX == "ITM") q = -seq(1,PERMLEVEL) else
                      if (x$moneyX == "OTM") q = seq(1,PERMLEVEL) } )    )
  ##   ncol is actually length of trades;  nrow is combo
  print(optGrid)
  allCombos <- list()
  
  for ( ix in 1:nrow(optGrid)) {
       ##   build the optStrat
       ###  If also using Long ? Short for the covered - need to be careful
       ### need to add that leg which will be just at that price
       ### add leg as an entry start of row 
       ###  rows denote combinations; columns are the legs
       ###  For fiexd Strikes row is number of lists; and strikes are each 
       ###  i.e., fixedStrikes[[jx]]  & then use fixedStrikes[[jx]][jy]
       combo <- list()
       for ( jy in 1:ncol(optGrid)) {
          if (str_detect  (specStrategy$legs$trades[[jy]]$name, "(Long)|(Short)")) typ = substr(specStrategy$legs$trades[[jy]]$name,1,1) else
          typ = ifelse(str_detect  (specStrategy$legs$trades[[jy]]$name, "Call"),  "C", "P" )
          name = specStrategy$legs$trades[[jy]]$name
          mult = 1
          ##if (typ == "C" & (specStrategy$legs$trades[[jy]]$moneyX == "OTM")) mult = -1
          ###    For puts  its reversed
          if (typ == "P" ) mult = -1
          
          ## if fised Strikes must use inputs given, not mktPrice based
          if (!tradeDefaults$fixedVals)
             strk = round(mktPrice * (1 + mult * optGrid[ix,jy] * tradeDefaults$incr  ),0)
          else
             strk = fixedStrikes[[ix]][jy] 
          
          lotSize = ifelse( is.null(specStrategy$legs$trades[[jy]]$lotsize), 
                            tradeDefaults$lotsize, specStrategy$legs$trades[[jy]]$lotsize)
          ###  get the premium from the quote slice
          ###  For long/Short - prem=0; since its not per share
          ###  If calendar spread, will have been specified in leg
          ###  get that slice - add to indxDate previously set in Driver
          if (!is.null(specStrategy$legs$trades[[jy]]$calendar)) 
            ##  there is a calendar dependency - get the slice
            mthQuote = allQuotes %>% dplyr::filter (dtx == friDates[[indxDate + specStrategy$legs$trades[[jy]]$calendar]])
          else
            mthQuote = qSlice
          if ( typ %in% c("C", "P"))  {
              qRow =  mthQuote %>% dplyr::filter(type == typ & strike == strk)
              prem = (qRow$bid + qRow$ask)/2  }
          else
              prem = 0    
       
          combo[[jy]] = list(name = specStrategy$legs$trades[[jy]]$name, 
                             strike = strk, premium = prem, lotsize = specStrategy$legs$trades[[jy]]$lotsize)
               }
  allCombos[[ix]] <- combo
     }
  out1 = lapply(  allCombos, function(caseX) 
                    return( list(specs = caseX,
                                 payoff = calcOptData(optStrat = caseX, stratName = stratName, atExpiry = atExpiry) )
                    )
  )
  
}
##############################
###
##############################
############################################################
####  Using Yahoo finance to get options chains  
###   Will get quotes for SPY - can use any other ticker
###   Also several expiration dates going out say 18-24 months
############################################################


#######################
###   function to get quotes for several expiration dates
#######################

getOptions <- function(ticker, dates, startD = startD) {
  require(RCurl)
  require(stringr)
  library(jsonlite)
  library(dplyr)
  curl = RCurl::getCurlHandle( cookiejar = '' )
  ####   To get cookies AND the crumb to use subsequently
  ckie = "https://finance.yahoo.com/lookup?s=bananas"
  xyzM = RCurl::getURL( ckie, curl = curl )
  urlOpt = "https://query1.finance.yahoo.com/v7/finance/options/%TICKER%?date=%DATE%"    ##  for "2020-10-16"
  ###
  ###    Note that time needs to be given in unix datetime POSIXct  -  data has also the same
  ###    use todays date or a hard coded one say "2020-04-22"
  
  odata = list()
  for ( ix in seq_along(dates) ) {
    urlOPT =  gsub("%TICKER%", ticker, gsub("%START%", startD, gsub( "%DATE%", dates[ix],  urlOpt)) )
    dat = RCurl::getURL(urlOPT, curl = curl)
    if ( length(grep("No data found", dat) == 1) > 0  )  next()
    odata[[ix]]  = dat
  }
  return(odata)
}
###########################################################
###   using third friday for 2020 & 2021 convert to unix type date
###   Functionalize this - given 1) years to get
###   determine 3rd Fri (option expiries) & get all quotes
###   Return as list so caller can process as needed
##############################################################

getOptQuotes <- function( parmList)  {
  yrs = parmList$yrs
  ticker = parmList$ticker
  fri = parmList$fri
  if (parmList$source == "CBOE" ){
      cboeQuotes <- read.csv(paste0(thisdir, "/quotedata_SPY.txt"), skip=2, header=T, as.is = T)
      df1 = cboeQuotes[,1:12]; df1[,2] = "C"
      df2 = cboeQuotes[,13:22];df2[,1] = "P"
      df2 = data.frame(df1[,1], df2, df1[,12])
      
      colnames(df2) = colnames(df1)
      quoteCBOE = rbind(df1,df2)
      quoteCBOE$Expiration.Date = as.Date(quoteCBOE$Expiration.Date, format = "%m/%d/%Y")
      unq = unique(quoteCBOE$Expiration.Date); 
      commonDates = unq[which(unq %in% fri)]   ####   use this to pick out the Quotes
      allQuotes = quoteCBOE[(quoteCBOE$Expiration.Date %in% commonDates),]
      ###   rename col names to conform to the Yahoo one ( for now !!!)
      colnames(allQuotes) = c("dtx", "type",  "lastPrice" , "Net",   "bid", "ask",  "vol",  "impliedVolatilit",              
                              "Delta", "Gamma", "openInterest", "strike" )  
      ###   to get market Price from furst line of the downloaded CBOE file
      firstLine = readLines(paste0(thisdir, "/quotedata_SPY.txt"),n=1)
      mktPrice = gsub(",", "", str_extract(firstLine, "(,[0-9.]*,)"))
      return(list(allQuotes = allQuotes, mktPrice = mktPrice))
  }  else {
      
        unxdate = sapply(fri, function(dt) as.numeric(as.POSIXct(dt)) )
      ###   pass ticker to get quotes for all expiry dates needed
      ###   start off with first Fri   
      allOData = getOptions(ticker, unxdate, fri[1])
      mktPrice = fromJSON(allOData[[1]])$optionChain$result$quote$regularMarketPrice 
      allQuotes = parseQuotes(allOData)
      ## retList <- list(); retList$allQuotes = allQuotes;  retList$mktPrice = mktPrice
      return(list(allQuotes = allQuotes, mktPrice = mktPrice, rawData = allOData) )
  }
  #return ( list(allQuotes = allQuotes, mktPrice = fromJSON(allOData[[1]]$optionChain$result$quote$regularMarketPrice ) ) )
}

##############################
#####    Get all the Fridays to use
#############################
get3rdFridays <- function(yrs) {
  library(RcppBDT)
  fri = as.Date(unlist(lapply(yrs,  function(yr) as.Date(sapply(seq(1,12), 
                                                                function(x) getNthDayOfWeek(third, Fri, x, yr) )) ) ) )
  ###   remove all before current date
  fri = fri [(fri > as.Date(Sys.Date() ) )]
  fri = fri [(fri < as.Date(paste0(yrs[length(yrs)],"-12-31")))]   
  return(fri)
  }

##############################
#####    Option functions
#############################

optCalcs <- function () {

  setOptParms <- function() {  
      optParms <- list()
      optParms$type = "c"  
      optParms$strike = 295
      optParms$Current = 282
      optParms$rf = 0.0015
      optParms$yld = 0.04   #   div yield per year
      optParms$mean = mean(spyRets["2019-04-24::"])
      optParms$sd = sd(spyLogRetsAnn["2008::"])
  }
  
  GBSOption("c", optParms$Current, optParms$strike, 0.5,  optParms$rf, optParms$rf, optParms$sd )@price
  
  ##   ref Tsay Analysis of Time Series pg 299-303 
  ## if sd is annual, then deltaT must be in terms of annual also, i.e., 6 months = 0.5 etc., 
  ###   formula for CDF normal h+  and h-   this is for call 
  ## h+  
  ##   Find the CDF for N(0,1)  at:-->      (log(Current/Strike) + (rf + sigma^2/2)*( deltaT) )   / ( sigma * sqrt(deltaT) )
  ##   Find the CDF for N(0,1)  at:-->      ( sigma * sqrt(deltaT) )
  
  ###   need to be careful about this - cant use 365 instead must use say 250 for year and adjust accordingly
  ###   can use days but then make sure everything lines up
  deltaT = 0.5     ###   in fraction of yr since std dev / volatility is based on daily returns over a year!           
  
  ### The D1  &  D2 parms referenced in any textbook   
  ###   Note the xval & zval below are usually referred to the N(d1)  & N(d2)  
  ###   All parms needed may be passed in via a formatted list
  xval =  (log(optParms$Current/optParms$strike) + (optParms$rf + optParms$sd ^ 2/ 2)*( deltaT) )   / (optParms$sd * sqrt(deltaT) )
  zval =  xval - (optParms$sd * sqrt(deltaT ))
  
  ###   pnorm gives us the CDF right away
  ####   no dividend s
  ## so for call 
  premCall = optParms$Current * pnorm(xval)  - optParms$strike * exp(-optParms$rf * deltaT) * pnorm(zval)
    ## so for put 
  premPut  = optParms$strike * exp(-optParms$rf * deltaT) * (1 - pnorm(zval)) - optParms$Current * (1 - pnorm(xval)) 
  
  #########################################
  ###    for dividend paying stock/index
  premCall = optParms$Current * exp(-optParms$yld * deltaT) * pnorm(xval)  - optParms$strike * exp(-optParms$rf * deltaT) * pnorm(zval)
  premPut  = optParms$strike * exp(-optParms$rf * deltaT) * (1 - pnorm(zval)) - optParms$Current *  exp(-optParms$yld * deltaT) * (1 - pnorm(xval)) 
  ###   above can also give us probabililties
  ### and for the potential movement of prices - that follow wiener process see Tsay pg 297
  ##  2500 * exp(0.3 * 0.5) * (exp(0.16 * 0.5) -1)     ???   formula price squared 
  ##1  price  at point deltaT in future;   and variance at that point
  
  #####################
  fcastPrice = optParms$Current *  exp ( optParms$mean * deltaT)
  fcastVar   = optParms$Current ^ 2 * exp( 2 * optParms$mean * deltaT) * (exp(optParms$sd ^2 * deltaT) -1 )
  ###   then std dev for this deltaT period will be
  sqrt(fcastVar)   ##   can now be used to get the probability cones for prices at time deltaT out in future!!!!
  ##############################
  #####    Option functions
  #############################
  
  
}
##############################
#####    Option functions
#############################

#############################
parseQuotes <- function (allOData) {
  #showCols = c( "contractSymbol", "strike", "openInterest", "lastPrice"  , "bid" ,  "ask", "percentChange", "impliedVolatility", "inTheMoney" )
  showCols = c( "strike", "openInterest", "lastPrice", "bid" ,  "ask", "impliedVolatility" )
  mypatt = '([A-Z]+)([0-9]{6})(C|P).*'
  
  fromDate = Sys.Date()
  
  xlst =   lapply (allOData, function(ixQuote) { 
    optFromJson = fromJSON(ixQuote)
    dfx =   optFromJson$optionChain$result$options
    dfx1 = dfx[[1]]
    print(dfx1$expirationDate)
    qCalls = dfx1$calls[[1]]  ####  this is finally the data frame of all calls in that expiry date
    qPuts  = dfx1$puts[[1]] 
    if (length(qCalls) == 0 )  return(NULL)
    ###   make sure all columns present
    c1 = which(!(showCols %in% colnames(qCalls)))
    c2 = which(!(showCols %in% colnames(qPuts)))
    if (!(length(c2) == 0  & length(c1) ==0) )  return(NULL)
    
    matched = str_match_all(qCalls$contractSymbol[1], mypatt)
    dtx = as.Date(paste0("20", matched[[1]][3]), format="%Y%m%d")
    ###   setup as dplyr table
    daysToExpiry = as.integer(difftime (as.Date(as.POSIXct(dfx1$expirationDate, origin="1970-01-01")) , fromDate))
    tblComb = as.tbl(cbind(dtx,
                           cbind( daysToExpiry, rbind( cbind( type="C",  qCalls[, showCols]), cbind(type="P", qPuts[, showCols])))) )
    
    return( tblComb)
  }  )
  
  allQuotes = do.call(rbind, xlst)
  return(allQuotes)
} 






###############################
